#! /bin/sh

CROSS="y"

# Settings for cross-compilation
cross_settings() {
    TARGET=x86/radios
    HOSTFILE=etc/radios-cross.c
    BUILDDIR=/opt/build/lcc-radios-cross
    LCCDIR=/opt/cross/radios/lcc
    
    OBJ_SUFFIX=".o"
    LIB_SUFFIX=".a"
    EXE_SUFFIX=""
    
    CC="cc"
    LD="cc"
    AR="ar ruvs"

    CFLAGS="-g -DLCCDIR=\\\"$LCCDIR/\\\" -DNDEBUG"
    LDFLAGS="-g"
}

# Settings for native compilation
native_settings() {
    TARGET=x86/radios
    HOSTFILE=etc/radios-native.c
    BUILDDIR=/opt/build/lcc-radios-native
    LCCDIR=/usr/local/lib/lcc
    
    OBJ_SUFFIX=".rdf"
    LIB_SUFFIX=".rdl"
    EXE_SUFFIX=".rdx"
    
    CC="ncc"
    LD="ncc"
    AR="rdlar ruvs"

    CFLAGS="-g -DLCCDIR=\\\"$LCCDIR/\\\" -DNDEBUG"
    LDFLAGS=""
}

# Prompt and read string from stdin
ReadData() {
    echo -n $3 "[$2]: "
    read p
    [ $p ] && eval $1=$p
}

# Replace @foo@ parameters with their real values
substitute() {
    [ "$2" != "=" ] && {
	echo $*
	return
    }
    printf "$1 $2 "
    case $3 in
	@TARGET@)	echo $TARGET		;;
	@HOSTFILE@)	echo $HOSTFILE		;;
	@BUILDDIR@)	echo $BUILDDIR		;;
	@LCCDIR@)	echo $LCCDIR		;;

	@OBJ_SUFFIX@)	echo $OBJ_SUFFIX	;;
	@LIB_SUFFIX@)	echo $LIB_SUFFIX	;;
	@EXE_SUFFIX@)	echo $EXE_SUFFIX	;;

	@CC@)		echo $CC		;;
	@LD@)		echo $LD		;;
	@AR@)		echo $AR		;;

	@CFLAGS@)	echo $CFLAGS		;;
	@LDFLAGS@)	echo $LDFLAGS		;;

	*)		echo $3			;;
    esac
}

# Main
ReadData CROSS $CROSS "Build cross-compiler?"
if [ "$CROSS" = "y" ] ; then
    cross_settings;
else
    native_settings;
fi

ReadData TARGET "$TARGET" "Target"
ReadData HOSTFILE "$HOSTFILE" "Host file"
ReadData BUILDDIR "$BUILDDIR" "Build directory"
ReadData LCCDIR "$LCCDIR" "LCC directory location"
echo
ReadData OBJ_SUFFIX "$OBJ_SUFFIX" "Object file suffix"
ReadData LIB_SUFFIX "$LIB_SUFFIX" "Library file suffix"
ReadData EXE_SUFFIX "$EXE_SUFFIX" "Executable file suffix"
echo
ReadData CC $CC "C compiler"
ReadData LD $LD "Linker"
ReadData AR "$AR" "Library archiver"
echo
ReadData CFLAGS "$CFLAGS" "C compiler parameters"
ReadData LDFLAGS "$LDFLAGS" "Linker parameters"
echo

echo '# Automatically generated file - do not edit. ' > header.mk
while read line; do
    substitute $line
done < header.mk.in >> header.mk

echo "Configuration file has been generated. Now you can run 'make'"
