/* ndk - [ pager.h ]
 *
 * Functions for interacting with the x86 pager
 *
 * (c)2004 dcipher / neuraldk
 *           www.neuraldk.org
 */

/**
* @defgroup pager pager.h: Intel pager functions
* @ingroup i386
*
 * Contains functions for working with the pager
*/
/** @{ */

#ifndef __ndk_pager_h__
#define __ndk_pager_h__

#include "console.h"
#include "errorCodes.h"
#include "types.h"

#define CR0_PAGING_ENABLE	(0x80000000)

// available for use in page directories and tables...
#define PAGE_PRESENT			(1)
#define PAGE_READ_WRITE		(2)
#define PAGE_P3_ACCESS		(4)
#define PAGE_WRITE_THROUGH	(8)
#define PAGE_CACHE_DISABLE	(16)
#define PAGE_ACCESSED		(32)

// page directories only
#define PAGE_SIZE_4MB		(128)

// page tables only
#define PAGE_DIRTY			(64)

// TODO: the following defines are from the old memory.c... clean them up!
#define PAGE_SIZE (4096)

// start allocating at 2GB (linear)
#define ALLOC_OFFSET     (524288)

#define MULTI_PAGE_ENTRY (1)
#define TOP_OF_DMA_STACK (4*1024*1024)

typedef struct
{
   long pageTable[1024];
} __attribute__ ((packed)) PageDirectory;

typedef struct
{
   long pageAddr[1024];
} __attribute__ ((packed)) PageTable;

typedef void *PhysicalAddress;
typedef void *LinearAddress;

ErrorCode pagerInit(void);
ErrorCode pagerGetPageSize(uint32 *page);
ErrorCode pagerGetNewPhysical(PhysicalAddress * addr);
ErrorCode pagerGetNewLinear(LinearAddress * addr);
ErrorCode pagerMapPhysicalToLinear(PhysicalAddress phys, LinearAddress lin);
ErrorCode pagerGetPhysical(LinearAddress lin, PhysicalAddress * addr);
ErrorCode pagerGetLinear(PhysicalAddress phys, LinearAddress * addr);   // possible/needed?

// remove/change
//ErrorCode pagerMapPage(unsigned long physAddr);
//void *pagerLinToPhys(void *linAddr);

#endif

/** @} */
